#!/usr/bin/perl -w

#
# dbmerge.pm
# Copyright (C) 1991-2013 by John Heidemann <johnh@isi.edu>
# $Id$
#
# This program is distributed under terms of the GNU general
# public license, version 2.  See the file COPYING
# in $dblibdir for details.
#

package Fsdb::Filter::dbmerge;

=head1 NAME

dbmerge - merge all inputs in sorted order based on the the specified columns

=head1 SYNOPSIS

    dbmerge --input A.fsdb --input B.fsdb [-T TemporaryDirectory] [-nNrR] column [column...]

or
    cat A.fsdb | dbmerge --input - --input B.fsdb [-T TemporaryDirectory] [-nNrR] column [column...]


or
    dbmerge [-T TemporaryDirectory] [-nNrR] column [column...] --inputs A.fsdb [B.fsdb ...]


=head1 DESCRIPTION

Merge all provided, pre-sorted input files, producing one sorted result.
Inputs can both be specified with C<--input>, or one can come
from standard input and the other from C<--input>.

Inputs must have identical schemas (columns, column order,
and field separators).

Unlike F<dbmerge2>, F<dbmerge> supports an arbitrary number of 
input files.

Because this program is intended to merge multiple sources,
it does I<not> default to reading from standard input.
If you wish to list F<-> as an explicit input source.

Also, because we deal with multiple input files,
this module doesn't output anything until it's run.

L<dbmerge> consumes a fixed amount of memory regardless of input size.
It therefore buffers output on disk as necessary.
(Merging is implemented a series of two-way merges,
so disk space is O(number of records).)

L<dbmerge> will merge data in parallel, if possible.
The <--parallelism> option can control the degree of parallelism,
if desired.


=head1 OPTIONS

General option:

=over 4

=item <--removeinputs>

Delete the source files after they have been consumed.
(Defaults off, leaving the inputs in place.)

=item <-T TmpDir>

where to put tmp files.
Also uses environment variable TMPDIR, if -T is 
not specified.
Default is /tmp.

=item <--parallelism> N

Allow up to N merges to happen in parallel.
Default is the number of CPUs in the machine.

=back

Sort specification options (can be interspersed with column names):

=over 4

=item B<-r> or B<--descending>

sort in reverse order (high to low)

=item B<-R> or B<--ascending>

sort in normal order (low to high)

=item B<-n> or B<--numeric>

sort numerically

=item B<-N> or B<--lexical>

sort lexicographically

=back

=for comment
begin_standard_fsdb_options

This module also supports the standard fsdb options:

=over 4

=item B<-d>

Enable debugging output.

=item B<-i> or B<--input> InputSource

Read from InputSource, typically a file name, or C<-> for standard input,
or (if in Perl) a IO::Handle, Fsdb::IO or Fsdb::BoundedQueue objects.

=item B<-o> or B<--output> OutputDestination

Write to OutputDestination, typically a file name, or C<-> for standard output,
or (if in Perl) a IO::Handle, Fsdb::IO or Fsdb::BoundedQueue objects.

=item B<--autorun> or B<--noautorun>

By default, programs process automatically,
but Fsdb::Filter objects in Perl do not run until you invoke
the run() method.
The C<--(no)autorun> option controls that behavior within Perl.

=item B<--help>

Show help.

=item B<--man>

Show full manual.

=back

=for comment
end_standard_fsdb_options


=head1 SAMPLE USAGE

=head2 Input:

File F<a.fsdb>:

    #fsdb cid cname
    11 numanal
    10 pascal

File F<b.fsdb>:

    #fsdb cid cname
    12 os
    13 statistics

These two files are both sorted by C<cname>,
and they have identical schemas.

=head2 Command:

    dbmerge --input a.fsdb --input b.fsdb cname

or

    cat a.fsdb | dbmerge --input b.fsdb cname

=head2 Output:

    #fsdb      cid     cname
    11 numanal
    12 os
    10 pascal
    13 statistics
    #  | dbmerge --input a.fsdb --input b.fsdb cname

=head1 SEE ALSO

L<dbmerge2(1)>,
L<dbsort(1)>,
L<Fsdb(3)>

=head1 CLASS FUNCTIONS

=cut


@ISA = qw(Fsdb::Filter);
($VERSION) = 2.0;

use strict;
use Carp qw(croak);
use Pod::Usage;

use threads;
use Thread::Semaphore;

use Fsdb::Filter;
use Fsdb::Filter::dbmerge2;
use Fsdb::IO::Reader;
use Fsdb::IO::Writer;
use Fsdb::Support::NamedTmpfile;
use Fsdb::Support::OS;


=head2 new

    $filter = new Fsdb::Filter::dbmerge(@arguments);

Create a new object, taking command-line arugments.

=cut

sub new ($@) {
    my $class = shift @_;
    my $self = $class->SUPER::new(@_);
    bless $self, $class;
    $self->set_defaults;
    $self->parse_options(@_);
    $self->SUPER::post_new();
    return $self;
}


=head2 set_defaults

    $filter->set_defaults();

Internal: set up defaults.

=cut

sub set_defaults ($) {
    my $self = shift @_;
    $self->SUPER::set_defaults();
    $self->{_remove_inputs} = undef;
    $self->{_info}{input_count} = 2;
    $self->{_sort_argv} = [];
    $self->{_max_parallelism} = undef;
    $self->{_test} = '';
    $self->set_default_tmpdir;
}

=head2 parse_options

    $filter->parse_options(@ARGV);

Internal: parse command-line arguments.

=cut

sub parse_options ($@) {
    my $self = shift @_;

    my(@argv) = @_;
    my $past_sort_options = undef;
    $self->get_options(
	\@argv,
 	'help|?' => sub { pod2usage(1); },
	'man' => sub { pod2usage(-verbose => 2); },
	'autorun!' => \$self->{_autorun},
	'close!' => \$self->{_close},
	'd|debug+' => \$self->{_debug},
	'i|input=s@' => sub { $self->parse_io_option('inputs', @_); },
	'inputs!' => \$past_sort_options,
	'log!' => \$self->{_logprog},
	'o|output=s' => sub { $self->parse_io_option('output', @_); },
	'parallelism=i' => \$self->{_max_parallelism},
	'removeinputs!' => \$self->{_remove_inputs},
	'test=s' => \$self->{_test},
	'T|tmpdir|tempdir=s' => \$self->{_tmpdir},
	# sort key options:
	'n|numeric' => sub { $self->parse_sort_option(@_); },
	'N|lexical' => sub { $self->parse_sort_option(@_); },
	'r|descending' => sub { $self->parse_sort_option(@_); },
	'R|ascending' => sub { $self->parse_sort_option(@_); },
	'<>' => sub { if ($past_sort_options) {
			    $self->parse_io_option('inputs', @_);
			} else  {
			    $self->parse_sort_option('<>', @_);
			};
		    },
	) or pod2usage(2);
}


=head2 segment_next_output

    $out = $self->segment_next_output($input_finished)

Internal: return a Fsdb::IO::Writer as $OUT
that either points to our output or a temporary file, 
depending on how things are going.

=cut

sub segment_next_output {
    my ($self, $input_finished) = @_;
    my $is_final_output = ($#{$self->{_files_to_merge}} == -1 && $input_finished);
    my $out;
    if ($is_final_output) {
#        $self->finish_io_option('output', -clone => $self->{_two_ins}[0]);
#        $out = $self->{_out};
	$out = $self->{_output};   # will pass this to the dbmerge2 module
	print "# final output\n" if ($self->{_debug});
    } else {
	# dump to a file for merging
	my $tmpfile = Fsdb::Support::NamedTmpfile::alloc($self->{_tmpdir});
	$out = $tmpfile;   # just return the name
    };
    return ($out, $is_final_output);
}


=head2 segment_cleanup

    $out = $self->segment_cleanup($file);

Internal: Clean up a file, if necessary.
(Sigh, used to be function pointers, but 
not clear how they would interact with threads.)

=cut

sub segment_cleanup($$) {
    my($self, $file) = @_;
    my($cleanup_type) = $self->{_files_cleanup}{$file};
    die "bad (empty) file in dbmerge::segment_cleanup\n"
	if (!defined($file));
    if (!defined($cleanup_type)) {
	# nothing
    } elsif ($cleanup_type eq 'unlink') {
	print "# cleaning up $file\n" if ($self->{_debug});
	unlink($file);
    } elsif ($cleanup_type eq 'NamedTmpfile') {
	Fsdb::Support::NamedTmpfile::cleanup_one($file);
    } else {
	die $self->{_prog} . ": internal error, unknown segment_cleanup type $cleanup_type\n";
    };
}


=head2 segments_finish_merge2

    $out = $self->segments_finish_merge2($out_fn, $is_final_output, 
			$two_in_fn_ref);


Internal: do the actual merge2 work (maybe our parent put us in a
thread, maybe not).

=cut
sub segments_finish_merge2($$$$) {
    my($self, $out_fn, $is_final_output, $two_in_fn_ref) = @_;

    my @merge_options = qw(--autorun --nolog);
    push(@merge_options, '--noclose', '--saveoutput' => \$self->{_out})
	if ($is_final_output);

    if ($self->{_debug}) {
	print "# merging $two_in_fn_ref->[0] with $two_in_fn_ref->[1] to $out_fn" . ($is_final_output ? " (final)" : "") . "\n";
    };
    new Fsdb::Filter::dbmerge2(@merge_options,
			'--input' => $two_in_fn_ref->[0],
			'--input' => $two_in_fn_ref->[1],
			'--output' => $out_fn,
			@{$self->{_sort_argv}});

    foreach (0..1) {
	$self->segment_cleanup($two_in_fn_ref->[$_]);
    };
}

=head2 segments_start_merge2

    $out = $self->segments_start_merge2($out_fn, $is_final_output, 
			$two_in_fn_ref);

Internal: handle all setup to merge2,
then do it in-line (if final) or in a thread (if not).

This routine runs in the main thread.
It is also the only place we control parallelism.

=cut

# Track parallelism here.
# it's shared across all instances of dbmerge, but that's a feature.
my $parallelism_available : shared = undef;

sub segments_start_merge2($$$$) {
    my($self, $out_fn, $is_final_output, $two_in_fn_ref) = @_;

    #
    # If input was from a forked thread,
    # reap it when it is done.
    #
    foreach my $i (0..1) {
	my $file = $two_in_fn_ref->[$i];
	next if (!ref($file));
	die "unkown object in merge queue.\n"
	    if (ref($file) ne 'ARRAY');
	# make sure they're running
	if (defined($file->[0])) {
	    print "# urgent start for thread\n" if ($self->{_debug});
	    $file->[0]->up();
	    $file->[0] = undef;
	};
	print "# waiting on thread\n" if ($self->{_debug});
	my $out_fn = $file->[1]->join;
	# substitute the now complete output file
	print "# intermediate file finished: $out_fn\n" if ($self->{_debug});
	$two_in_fn_ref->[$i] = $out_fn;
    };

    # Take files in order with the goal of providing a stable sort,
    # (if they're files... do nothing if they're objects).
    if (!ref($two_in_fn_ref->[0]) && !ref($two_in_fn_ref->[1])) {
	if (($two_in_fn_ref->[0] cmp $two_in_fn_ref->[1]) > 0) {
	    @{$two_in_fn_ref}[1,0] = @{$two_in_fn_ref}[0,1];
	};
    };

    if ($is_final_output) {
        $self->segments_finish_merge2($out_fn, $is_final_output, $two_in_fn_ref);
    } else {
	print "# intermediate file started: $out_fn\n" if ($self->{_debug});
	$self->{_files_cleanup}{$out_fn} = 'NamedTmpfile';
	#
	# Synchronize thread start so we can guarantee
	# we're ready to reap it before it starts and maybe finishes.
	# 
	my $go_sem = Thread::Semaphore->new(0);
	my $merge_thread = threads->new(
	    sub {
		$go_sem->down();  # wait to be told to start
		$parallelism_available--;
		$self->segments_finish_merge2($out_fn, $is_final_output, $two_in_fn_ref);
		sleep(1) if (defined($self->{_test}) && $self->{_test} eq 'delay');
		$parallelism_available++;
		return $out_fn;
	    }
        );
	# put the thread in our queue, which will cause it to be reaped
	push(@{$self->{_files_to_merge}}, [$go_sem, $merge_thread]);
	#
	# increase parallelism, if possible.
	#
	my $threads_to_release = $parallelism_available;
	print "# thread starting, with up to $threads_to_release\n" if ($self->{_debug});
	foreach my $file (@{$self->{_files_to_merge}}) {
	    if ($threads_to_release <= 0) {
		last;
	    };
	    next if (!defined(ref($file)) || ref($file) eq '');   # filename
	    die "unknown entry in _files_to_merge: " . ref($file) . "\n"
		if (ref($file) ne 'ARRAY');
	    if (defined($file->[0])) {
		# found a pending task; start it
		print "# releasing thread\n" if ($self->{_debug});
		$file->[0]->up();
		$file->[0] = undef;
		$threads_to_release--;
	    };
	};
    };
}

=head2 segments_merge_all

    $self->segments_merge_all();

Merge queued files, if any.

We process the work queue in
a file-system-cache-friendly order, based on ideas from
"Information and Control in Gray-box Systems" by
the Arpaci-Dusseau's at SOSP 2001.

Idea:  each "pass" through the queue, reverse the processing
order so that the most recent data (that's hopefully
in the file system cache in memory) is handled first.

This algorithm isn't perfect (sometimes if there's an odd number
of files in the queue you reach way back in time, but most of 
the time it's quite good).

=cut

sub segments_merge_all ($) {
    my($self) = @_;
    return if ($#{$self->{_files_to_merge}} == -1);

    my $files_to_merge_ref = $self->{_files_to_merge};

    # keep track of how often to reverse
    my($files_before_reversing_queue) = $#{$files_to_merge_ref} + 1;
    # Merge the files in a binary tree.
    while ($#{$files_to_merge_ref} >= 0) {
	# Each "pass", reverse the queue to reuse stuff in memory.
	if ($files_before_reversing_queue <= 0) {
	    @{$files_to_merge_ref} = reverse @{$files_to_merge_ref};
	    $files_before_reversing_queue = $#{$files_to_merge_ref} + 1;
	    print "# reversed queue, $files_before_reversing_queue before next reverse.\n" if ($self->{_debug});
	};
	$files_before_reversing_queue -= 2;

	# Pick off the two next files for merging.
	die $self->{_prog} . ": segments_merge_all, odd number of segments.\n" if ($#{$files_to_merge_ref} == 0);   # assert
	my (@two_fn) = (shift @${files_to_merge_ref}, shift @${files_to_merge_ref});
	my($out_fn, $is_final_output) = $self->segment_next_output(1);

	#
	# run our merge
	#
	$self->segments_start_merge2($out_fn, $is_final_output, \@two_fn);
    };
}


=head2 setup

    $filter->setup();

Internal: setup, parse headers.

=cut

sub setup ($) {
    my($self) = @_;

    croak $self->{_prog} . ": no sorting key specified.\n"
	if ($#{$self->{_sort_argv}} == -1);

    if ($#{$self->{_inputs}} == -1) {
	croak $self->{_prog} . ": no input sources specified, use --input.\n";
    };
    @{$self->{_files_to_merge}} = @{$self->{_inputs}};
    # prove files exist (early error checking)
    foreach (@{$self->{_inputs}}) {
	next if (ref($_) ne '');   # skip objects
	next if ($_ eq '-');   # special case: stdin
	if (! -f $_) {
	    croak $self->{_prog} . ": input source $_ does not exist.\n";
	};
    };
    # don't
    if ($self->{_remove_inputs}) {
	foreach (@{$self->{_files_to_merge}}) {
	    $self->{_files_cleanup}{$_} = 'unlink';
	};
    };
    #
    # control parallelism
    #
    if (!defined($parallelism_available)) {
        if (!defined($self->{_max_parallelism})) {
	    $self->{_max_parallelism} = Fsdb::Support::OS::max_parallelism();
	};
	$parallelism_available = $self->{_max_parallelism};
    };
}

=head2 run

    $filter->run();

Internal: run over each rows.

=cut
sub run ($) {
    my($self) = @_;

    $self->segments_merge_all();
};

    

=head1 AUTHOR and COPYRIGHT

Copyright (C) 1991-2013 by John Heidemann <johnh@isi.edu>

This program is distributed under terms of the GNU general
public license, version 2.  See the file COPYING
with the distribution for details.

=cut

1;
